const g = require('../globals');
const ui = require('../ui');
const $ = require('jquery');
const ko = require('knockout');
const components = require('./components');
const bbgmView = require('../util/bbgmView');
const helpers = require('../util/helpers');

let listenersAdded = false;

/**
 * Generate a box score.
 *
 * @memberOf views.gameLog
 * @param {number} gid Integer game ID for the box score (a negative number means no box score).
 * @return {Promise.Object} Resolves to an object containing the box score data (or a blank object).
 */
async function boxScore(gid) {
    if (gid < 0) {
        return {};
    }

    const game = await g.dbl.games.get(gid);

    // If game doesn't exist (bad gid or deleted box scores), show nothing
    if (!game) {
        return {};
    }

    for (let i = 0; i < game.teams.length; i++) {
        const t = game.teams[i];

        // Team metadata
        t.abbrev = g.teamAbbrevsCache[t.tid];
        t.region = g.teamRegionsCache[t.tid];
        t.name = g.teamNamesCache[t.tid];

        // four factors
        t.efg = 100 * (t.fg + (t.tp / 2)) / t.fga;
        t.tovp = 100 * t.tov / (t.fga + 0.44 * t.fta + t.tov);
        t.orbp = 100 * t.orb / (t.orb + game.teams[1 - i].drb);
        t.ftpfga = t.ft / t.fga;

        // Fix the total minutes calculation, which is usually fucked up for some unknown reason
        t.min = 240 + 25 * game.overtimes;

        // Put injured players at the bottom, then sort by GS and roster position
        t.players.sort((a, b) => {
            // This sorts by starters first and minutes second, since .min is always far less than 1000 and gs is either 1 or 0. Then injured players are listed at the end, if they didn't play.
            return (b.gs * 100000 + b.min * 1000 - b.injury.gamesRemaining) - (a.gs * 100000 + a.min * 1000 - a.injury.gamesRemaining);
        });
    }

    // Team metadata
    game.won.region = g.teamRegionsCache[game.won.tid];
    game.won.name = g.teamNamesCache[game.won.tid];
    game.won.abbrev = g.teamAbbrevsCache[game.won.tid];
    game.lost.region = g.teamRegionsCache[game.lost.tid];
    game.lost.name = g.teamNamesCache[game.lost.tid];
    game.lost.abbrev = g.teamAbbrevsCache[game.lost.tid];

    if (game.overtimes === 1) {
        game.overtime = " (OT)";
    } else if (game.overtimes > 1) {
        game.overtime = ` (${game.overtimes}OT)`;
    } else {
        game.overtime = "";
    }

    // Quarter/overtime labels
    game.qtrs = ["Q1", "Q2", "Q3", "Q4"];
    for (let i = 0; i < game.teams[1].ptsQtrs.length - 4; i++) {
        game.qtrs.push(`OT${i + 1}`);
    }
    game.qtrs.push("F");

    return game;
}

function get(req) {
    return {
        abbrev: helpers.validateAbbrev(req.params.abbrev)[1],
        gid: req.params.gid !== undefined ? parseInt(req.params.gid, 10) : -1,
        season: helpers.validateSeason(req.params.season),
    };
}

function InitViewModel() {
    this.boxScore = {
        gid: ko.observable(-1),
        prevGid: ko.observable(null),
        nextGid: ko.observable(null),
    };
    this.gamesList = {
        abbrev: ko.observable(),
        loading: ko.observable(true), // Needed because this isn't really set until updateGamesList, which could be after first render
        season: ko.observable(),
        games: ko.observableArray([]),
    };

    // This computed is used so the box score won't be rendered until after it is fully loaded (due to the throttle). Otherwise, the mapping plugin sometimes sets the gid before the rest of the box score.
    // But because it's throttled, ui.tableClickableRows can't be called directly in uiFirst or uiEvery.
    this.showBoxScore = ko.computed(function () {
        return this.boxScore.gid() >= 0;
    }, this).extend({throttle: 1});
}

/* This doesn't work for some reason.
const mapping = {
    gamesList: {
        update(options) {
            return new function () {
                komapping.fromJS(options.data, {
                    games: {
                        create: options => options.data;
                    }
                }, this);
            }();
        }
    }
};*/

function updatePrevNextLinks(vm) {
    const games = vm.gamesList.games();
    vm.boxScore.prevGid(null);
    vm.boxScore.nextGid(null);

    for (let i = 0; i < games.length; i++) {
        if (games[i].gid === vm.boxScore.gid()) {
            if (i > 0) {
                vm.boxScore.nextGid(games[i - 1].gid);
            }
            if (i < games.length - 1) {
                vm.boxScore.prevGid(games[i + 1].gid);
            }
            break;
        }
    }
}

function updateTeamSeason(inputs) {
    return {
        // Needed for dropdown
        abbrev: inputs.abbrev,
        season: inputs.season,
    };
}

/**
 * Update the displayed box score, as necessary.
 *
 * If the box score is already loaded, nothing is done.
 *
 * @memberOf views.gameLog
 * @param {number} inputs.gid Integer game ID for the box score (a negative number means no box score).
 */
async function updateBoxScore(inputs, updateEvents, vm) {
    if (updateEvents.indexOf("dbChange") >= 0 || updateEvents.indexOf("firstRun") >= 0 || inputs.gid !== vm.boxScore.gid()) {
        const game = await boxScore(inputs.gid);

        const vars = {
            boxScore: game,
        };

        // Either update the box score if we found one, or show placeholder
        if (!game.hasOwnProperty("teams")) {
            vars.boxScore.gid = -1;
        } else {
            vars.boxScore.gid = inputs.gid;

            // Force scroll to top, which otherwise wouldn't happen because this is an internal link
            window.scrollTo(window.pageXOffset, 0);
        }

        return vars;
    }
}

/**
 * Update the game log list, as necessary.
 *
 * If the game log list is already loaded, nothing is done. If the game log list is loaded and a new game has been played, update. If the game log list is not loaded, load it.
 *
 * @memberOf views.gameLog
 * @param {string} inputs.abbrev Abbrev of the team for the list of games.
 * @param {number} inputs.season Season for the list of games.
 * @param {number} inputs.gid Integer game ID for the box score (a negative number means no box score), which is used only for highlighting the relevant entry in the list.
 */
async function updateGamesList(inputs, updateEvents, vm) {
    if (updateEvents.indexOf("dbChange") >= 0 || updateEvents.indexOf("firstRun") >= 0 || inputs.abbrev !== vm.gamesList.abbrev() || inputs.season !== vm.gamesList.season()) {
        // Load all games in list
        vm.gamesList.loading(true);
        vm.gamesList.games([]);
        const games = await helpers.gameLogList(inputs.abbrev, inputs.season, inputs.gid, vm.gamesList.games());
        vm.gamesList.games(games);
        vm.gamesList.abbrev(inputs.abbrev);
        vm.gamesList.season(inputs.season);
        vm.gamesList.loading(false);

        // Update prev/next links, in case box score loaded before games list
        updatePrevNextLinks(vm);

/* This doesn't work for some reason.
        return {
            gamesList: {
                games: games,
                abbrev: inputs.abbrev,
                season: inputs.season,
                loading: false
            }
        };*/
    }
    if (updateEvents.indexOf("gameSim") >= 0 && inputs.season === g.season) {
        // Partial update of only new games
        const games = await helpers.gameLogList(inputs.abbrev, inputs.season, inputs.gid, vm.gamesList.games());
        for (let i = games.length - 1; i >= 0; i--) {
            vm.gamesList.games.unshift(games[i]);
        }

        // Update prev/next links, in case box score loaded before games list
        updatePrevNextLinks(vm);
    }
}

function uiFirst(vm) {
    ko.computed(() => {
        ui.title(`Game Log - ${vm.season()}`);
    }).extend({throttle: 1});

    // Update prev/next links whenever box score gid is changed
    ko.computed(() => {
        vm.boxScore.gid();
        updatePrevNextLinks(vm);
    }).extend({throttle: 1});

    // Would be better if I had an "unmount" function...
    if (!listenersAdded) {
        listenersAdded = true;
        document.addEventListener("keydown", e => {
            let el;
            if (e.keyCode === 37) {
                el = document.getElementById('game-log-prev');
            } else if (e.keyCode === 39) {
                el = document.getElementById('game-log-next');
            }

            if (el) {
                el.click();
            }
        });
    }
}

function uiEvery(updateEvents, vm) {
    components.dropdown("game-log-dropdown", ["teams", "seasons"], [vm.abbrev(), vm.season()], updateEvents, vm.boxScore.gid() >= 0 ? vm.boxScore.gid() : undefined);

    // UGLY HACK for two reasons:
    // 1. Box score might be hidden if none is loaded, so in that case there is no table to make clickable
    // 2. When box scores are shown, it might happen after uiEvery is called because vm.showBoxScore is throttled
    window.setTimeout(() => {
        const tableEls = $(".box-score-team");
        if (tableEls.length > 0 && !tableEls[0].classList.contains("table-hover")) {
            ui.tableClickableRows(tableEls);
        }
    }, 100);
}

module.exports = bbgmView.init({
    id: "gameLog",
    get,
    InitViewModel,
    runBefore: [updateBoxScore, updateTeamSeason],
    runWhenever: [updateGamesList],
    uiFirst,
    uiEvery,
});
